/*
	dotNET FontManager library
    Copyright (C) 2015 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    This license has exceptions: it can be used with OxyPlot and SvgNet on basis 
    on Apache license. For all other cases it shoulkd be provided under GPLv3.
*/

using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Globalization;
using System.Runtime.InteropServices;
using System.Text;

namespace fonts.freetype
{
	public class GlyphInfo 
	{
		public GraphicsPath gp = new GraphicsPath();
		public PointF advance = new PointF(0, 0);
		public RectangleF bbox= new RectangleF(0,0,0,0);

		public float width  = 0;
		public float height = 0;
		public float shiftX = 0;
		public float shiftY = 0; 
		public float advm   = 0;

		public float fontAscent = 0;
		public float fontHeight = 0;
		public float fontDescent= 0;
		public float unitsPerEm = 0;
		public float lineSpacing= 0;

		public char cc;
		public int glyph;
	}

	public class FreetypeFontSupplier 
	{
		private Matrix at;
		private Hashtable knownGlyphs;
		private byte[] data;

		public float fontAscent = 0;
		public float fontHeight = 0;
		public float fontDescent= 0;
		public float unitsPerEm = 0;
		public float lineSpacing= 0;

		public int DFL_K = 1;

		public FreetypeFontSupplier (byte[] data, Matrix at) {
			this.data = data;
			knownGlyphs = new Hashtable();
			this.at = at;

			if(this.at == null) {
				DFL_K = 100;
				this.at = new Matrix (1f / DFL_K, 0, 0, 1f / DFL_K, 0, 0);
			}

			precache();
		}

		public int Count { get { return knownGlyphs == null ? 0 : knownGlyphs.Count; } }

		public GlyphInfo getGlyph(int code) {
    		return (GlyphInfo)knownGlyphs[code];
		}

		[DllImport("freetypewrap.dll", 
    		CallingConvention = CallingConvention.Winapi, 
    		CharSet = CharSet.Auto)]
		[return: MarshalAs(UnmanagedType.I1)]
		private static extern bool load(IntPtr inb, int inlen, out IntPtr arrayPtr, out int size);

		private static byte[] f_load(byte[] inb) {
			IntPtr outp = IntPtr.Zero;
			int outlen = 0;

			int inlen = Marshal.SizeOf(inb[0]) * inb.Length;
			IntPtr inp = Marshal.AllocHGlobal(inlen);
			try	{
				Marshal.Copy(inb, 0, inp, inb.Length);
				if(load(inp, inlen, out outp, out outlen)) {
					byte[] outb = new byte[outlen];
					Marshal.Copy(outp, outb, 0, outlen);
					return outb;
				}
			}
			catch(Exception e) {
				Console.Error.WriteLine("!!! Error loading freetypewrap.dll !!! " + e.Message);
			}
			finally {
				Marshal.FreeHGlobal(inp);
			}
			return null;
		}			

		private void precache() 
		{
    		byte[] retnb = f_load(data);
    		if(retnb == null) return;

			GlyphInfo gi = null;
    		int lastCode = 0;

			System.Text.Encoding enc = System.Text.Encoding.GetEncoding("UTF-8");
    		String retn = enc.GetString(retnb);

    		//System.err.println(retn);
    		retn = retn.Replace("\r", "\n");
    		PointF sp = new PointF(0, 0);
			bool added  = false;

			int ascent = 0, descent = 0, height = 0, em = 1000, max_ah = 0;

			String[] st = retn.Split(new char[]{'\n'});
    		for(int jjk = 0; jjk < st.Length; ++jjk) {
	    		String line = st[jjk].Trim();
	    		int j = line.IndexOf('#');
    			if(j > 0) line = line.Substring(0, j).Trim();
	    		if(line.StartsWith("#")) line = "";
	    		if(line.Length < 1) continue;

				JavaTokenizer sx = new JavaTokenizer(line, " \t");
				while(sx.hasMoreTokens()) {
	    			String dmd = sx.nextToken();

		   			if(dmd.Equals("G")) {
	    				gi = new GlyphInfo();
	    				int code = nextInt(sx);
    					float x = nextFloat(sx);
    					float y = nextFloat(sx);
    					int uucode = nextInt(sx);
	    				gi.advance = new PointF(x / em, y / em);
	    				gi.cc = (char)uucode;
	    				gi.glyph = code;
    					lastCode = uucode < 0 ? 0x40000000 + code : uucode;
    					knownGlyphs[lastCode] = gi;
    					sp = new PointF(0, 0);
    					added = false;

    					gi.fontAscent = ascent;
    					gi.fontHeight = height;
    					gi.fontDescent= descent;
    					gi.unitsPerEm = em;
    					gi.lineSpacing= max_ah;
    					
    				}
	    			else if(dmd.Equals("B")) {
    					float x1 = nextFloat(sx);
    					float y1 = nextFloat(sx);
    					float x2 = nextFloat(sx);
    					float y2 = nextFloat(sx);
    					gi.bbox = new RectangleF(x1 / (float)em, y1 / (float)em, x2 / (float)em, y2 / (float)em);
	    			}
	    			else if(dmd.Equals("F")) {	
    					gi.gp.FillMode = FillMode.Winding;
    				}
	    			else if(dmd.Equals("f")) {
	    				ascent = nextInt(sx);
	    				descent= nextInt(sx);
	    				height = nextInt(sx);
	    				em     = nextInt(sx) * DFL_K;
	    				max_ah = nextInt(sx);

    					this.fontAscent  = (float)ascent / (float)em;
    					this.fontDescent = (float)descent / (float)em;
    					this.fontHeight  = (float)height / (float)em;
    					this.unitsPerEm  = (float)em;
    					this.lineSpacing = (float)max_ah / (float)em;
	    			}
	    			else if(dmd.Equals("x")) {
	    				gi.width  = nextFloat(sx) / (float)em;
	    				gi.height = nextFloat(sx) / (float)em;
	    				gi.shiftX = nextFloat(sx) / (float)em;
	    				gi.advm   = nextFloat(sx) / (float)em;
	    				gi.shiftY = nextFloat(sx) / (float)em;
	    			}
	    			else if(dmd.Equals("g")) {
    					gi.gp.Transform(this.at);
    					gi = null;
    				}
	    			else if(dmd.Equals("s")) { 
    					if(added) gi.gp.CloseFigure();
    					added = false;
    				}
	    			else if(dmd.Equals("M")) {
    					float x = nextFloat(sx);
    					float y = nextFloat(sx);
    					sp = new PointF(x, y);
    				}
	    			else if(dmd.Equals("L")) {
    					float x = nextFloat(sx);
    					float y = nextFloat(sx);
    					gi.gp.AddLine(sp.X, sp.Y, x, y);
    					sp = new PointF(x, y);
    					added = true;
	    			}
	    			else if(dmd.Equals("C")) {
    					float x1 = nextFloat(sx);
    					float y1 = nextFloat(sx);
    					float x2 = nextFloat(sx);
    					float y2 = nextFloat(sx);
    					float x3 = nextFloat(sx);
	    				float y3 = nextFloat(sx);
    					gi.gp.AddBezier(sp.X, sp.Y, x1, y1, x2, y2, x3, y3);
    					sp = new PointF(x3, y3);
    					added = true;
    				}
	    			else if(dmd.Equals("Q")) {
	    				float x1 = nextFloat(sx);
    					float y1 = nextFloat(sx);
    					float x2 = nextFloat(sx);
    					float y2 = nextFloat(sx);

    					/*
							"quadratic bezier of the form [P1, C, P2]
							(where C is the control point) you can form an equivalent cubic bezier
							with [P1, (C*2/3 + P1 * 1/3), (C*2/3 + P2 * 1/3), P2]. "
	    					gi.gp.quadTo(x1, y1, x2, y2);
    					*/

    					gi.gp.AddBezier(sp.X, sp.Y, 
    						x1 * 2/3 + sp.X * 1/3, y1 * 2/3 + sp.Y * 1/3, 
    						x1 * 2/3 + x2   * 1/3, y1 * 2/3 + y2 * 1/3, 
    						x2, y2);
    					sp = new PointF(x2, y2);
    					added = true;
    				}
	    		}
    		}

    		// Debug
    		// Console.WriteLine(retn);
    		//Console.WriteLine("Font: ascent/descent=" + fontAscent + "/" + fontDescent + ", " + 
    		//	"height=" + fontHeight + ", spacing=" + lineSpacing + ", u/Em=" + unitsPerEm);
		}

		private float nextFloat(JavaTokenizer sx) {
    		String x = sx.nextToken();
			return Single.Parse(x, CultureInfo.InvariantCulture);
		}

		private int nextInt(JavaTokenizer sx) {
			String x = sx.nextToken();
			return Int32.Parse(x, CultureInfo.InvariantCulture);
		}

		// public API
		// Measurements

		private void computeW(float[] w, float[] maxW, char cc) 
		{
			int code  = (int)cc;
			GlyphInfo gi = getGlyph(code);
			if(cc == '\n' || cc == '\r') {
				maxW[0] = Math.Max(maxW[0], w[0]);
				w[0] = 0;
			}
			else if(gi != null) 
				w[0] += gi.advance.X;
		}

		public float MeasureTextWidth(String text, float emSize)
		{
		    float[] width = new float[1];
		    float[] maxW  = new float[1];
		    width[0] = maxW[0] = 0;

			for(int k = 0; k < text.Length; ++k) {
				char cc = text[k];
				computeW(width, maxW, cc);
			}
		    width[0] = Math.Max(maxW[0], width[0]);
			return width[0] * emSize;
		}

		public SizeF MeasureText(String text, float emSize)
		{
			SizeF sz = new SizeF();
			sz.Width = MeasureTextWidth(text, emSize);
			sz.Height = GetHeight(emSize);
        	for(int k = 0; k < text.Length; ++k) {
        	    if(text[k] == '\n' || text[k] == '\r') {
        	    	if(text[k] == '\n' && k + 1 < text.Length && text[k+1] == '\r') ++k;
        	    	if(text[k] == '\r' && k + 1 < text.Length && text[k+1] == '\n') ++k;
        	    	if(k < text.Length) sz.Height += GetHeight(emSize);
        	    }
        	}
        	return sz;
		}

		public float GetHeight(float emSize) 
		{
			return fontHeight * emSize;
		}

		public float GetAscent(float emSize) 
		{
			return fontAscent * emSize;
		}

        public float GetDescent(float emSize) 
        {
            return fontDescent * emSize;
        }

        // paint
        public void DrawString(Graphics g, PointF pt, String text, float emSize, Brush brush)
        {
        	DrawString(g, pt.X, pt.Y, -1, -1, text, emSize, brush);
        }

        public void DrawString(Graphics g, float x, float y, String text, float emSize, Brush brush)
        {
        	DrawString(g, x, y, -1, -1, text, emSize, brush);
        }

        public void DrawString(Graphics g, PointF pt, SizeF sz, String text, float emSize, Brush brush)
        {
        	DrawString(g, pt.X, pt.Y, sz.Width, sz.Height, text, emSize, brush);
        }

        public void DrawString(Graphics g, RectangleF r, String text, float emSize, Brush brush)
        {
        	DrawString(g, r.Left, r.Top, r.Right - r.Left, r.Bottom - r.Top, text, emSize, brush);
        }

        public void DrawString(Graphics g, float x, float y, float w, float h, String text, float emSize, Brush brush)
        {
        	float startX = x;
        	float startY = y + GetAscent(emSize);
        	for(int k = 0; k < text.Length; ++k) {
        	    if(text[k] == '\n' || text[k] == '\r' || (w > 0 && startX >= x + w)) {
        	    	if(text[k] == '\n' && k + 1 < text.Length && text[k+1] == '\r') ++k;
        	    	if(text[k] == '\r' && k + 1 < text.Length && text[k+1] == '\n') ++k;
        	    	startX = x;
        	    	startY += GetHeight(emSize);
        	    	if(h >0 && startY > y + h) break;
        	    	continue;
        	    }

        		GlyphInfo gi = getGlyph(text[k]);
        		if(gi == null) continue;
		    	GraphicsPath gp = (GraphicsPath)gi.gp.Clone();

		    	Matrix at = new Matrix();
		    	at.Translate(startX, startY);
		    	at.Scale(emSize, emSize);

		    	Matrix oldAT = g.Transform;
		    	g.MultiplyTransform(at);
		    	g.FillPath(brush, gp);
		    	g.Transform = oldAT;

		    	startX += gi.advance.X * emSize;
        	}
        }

        // paint outline
        public void DrawOutline(Graphics g, PointF pt, String text, float emSize, Pen pen)
        {
        	DrawOutline(g, pt.X, pt.Y, -1, -1, text, emSize, pen);
        }

        public void DrawOutline(Graphics g, float x, float y, String text, float emSize, Pen pen)
        {
        	DrawOutline(g, x, y, -1, -1, text, emSize, pen);
        }

        public void DrawOutline(Graphics g, PointF pt, SizeF sz, String text, float emSize, Pen pen)
        {
        	DrawOutline(g, pt.X, pt.Y, sz.Width, sz.Height, text, emSize, pen);
        }

        public void DrawOutline(Graphics g, RectangleF r, String text, float emSize, Pen pen)
        {
        	DrawOutline(g, r.Left, r.Top, r.Right - r.Left, r.Bottom - r.Top, text, emSize, pen);
        }

        public void DrawOutline(Graphics g, float x, float y, float w, float h, String text, float emSize, Pen pen)
        {
        	float startX = x;
        	float startY = y + GetAscent(emSize);
        	for(int k = 0; k < text.Length; ++k) {
        	    if(text[k] == '\n' || text[k] == '\r' || (w > 0 && startX >= x + w)) {
        	    	if(text[k] == '\n' && k + 1 < text.Length && text[k+1] == '\r') ++k;
        	    	if(text[k] == '\r' && k + 1 < text.Length && text[k+1] == '\n') ++k;
        	    	startX = x;
        	    	startY += GetHeight(emSize);
        	    	if(h >0 && startY > y + h) break;
        	    	continue;
        	    }

        		GlyphInfo gi = getGlyph(text[k]);
        		if(gi == null) continue;
		    	GraphicsPath gp = (GraphicsPath)gi.gp.Clone();

		    	Matrix at = new Matrix();
		    	at.Translate(startX, startY);
		    	at.Scale(emSize, emSize);

		    	Matrix oldAT = g.Transform;
		    	g.MultiplyTransform(at);
		    	g.DrawPath(pen, gp);
		    	g.Transform = oldAT;

		    	startX += gi.advance.X * emSize;
        	}
        }

        // SVG related

		public float UnitsPerEm { get { return unitsPerEm; } }
		public float GetAdvanceX(char c) {
	    	GlyphInfo gi = getGlyph(c);
	    	if(gi == null) return 0.0f;
	    	return gi.advance.X * UnitsPerEm;
		}

		public float MatrixScale {
			get { return this.at.Elements[0]; }
		}

    	public String ToSVGPathData(char c) 
    	{
			StringBuilder d = new StringBuilder();

	    	GlyphInfo gi = getGlyph(c);
	    	GraphicsPath gp;
	    	if(gi == null) gp = new GraphicsPath();
	    	else {
	    		gp = (GraphicsPath)gi.gp.Clone();
		        Matrix at = new Matrix();
    		    at.Scale(1.0f/this.at.Elements[0], 1.0f/this.at.Elements[3]);
        		gp.Transform(at);
            }
		
			float lastX = 0;
			float lastY = 0;
			
			int strIdx;
			int pathType;
			float x1, y1, x2, y2, x3, y3;
			int iptr;

			GraphicsPath s = gp;
			for (iptr = 0; iptr < s.PathData.Types.Length;)
			{
				pathType = (int)s.PathData.Types[iptr];
				strIdx = iptr;
				bool close = (pathType & (int)PathPointType.CloseSubpath) == (int)PathPointType.CloseSubpath;
				bool dashed= (pathType & (int)PathPointType.DashMode) == (int)PathPointType.DashMode;

				pathType &= (int)PathPointType.PathTypeMask;
				switch((PathPointType)pathType) 
				{
					/* PathPointType.Bezier3 == PathPointType.Bezier */
					case PathPointType.Bezier3:
						x1 = s.PathData.Points[strIdx].X;
						y1 = s.PathData.Points[strIdx].Y;
						x2 = s.PathData.Points[strIdx+1].X;
						y2 = s.PathData.Points[strIdx+1].Y;
						x3 = s.PathData.Points[strIdx+2].X;
						y3 = s.PathData.Points[strIdx+2].Y;
						lastX = x3;
						lastY = y3;
                		d.Append("C ");
               			appendPoint(d, x1, y1);
                		appendPoint(d, x2, y2);
                		appendPoint(d, x3, y3);
						iptr += 3;
						break;
					case PathPointType.Line:
						lastX = s.PathData.Points[strIdx].X;
						lastY = s.PathData.Points[strIdx].Y;
                		d.Append("L ");
                		appendPoint(d, lastX, lastY);
						++iptr;
						break;
					case PathPointType.Start: 
						lastX = s.PathData.Points[strIdx].X;
						lastY = s.PathData.Points[strIdx].Y;
                		d.Append("M ");
                		appendPoint(d, lastX, lastY);
						++iptr;
						break;
				}

				if(close)
					d.Append("Z ");
			}

        	if (d.Length > 0)
            	return d.ToString().Trim();
        	else // emty path
            	return "";
    	}

    	private void appendPoint(StringBuilder d, float x, float y) {
        	d.Append(x);
        	d.Append(' ');
        	d.Append(y);
        	d.Append(' ');
    	}

    	private void appendAttr(StringBuilder sb, String k, String v) {
    		sb.Append(' ');
    		sb.Append(k);
    		sb.Append('=');
    		sb.Append('\"');
    		sb.Append(v);
    		sb.Append('\"');
    	}
	}

	public class JavaTokenizer
	{
		private int currentPosition;
		private int newPosition;
		private int maxPosition;
		private String str;
		private String delimiters;
		private bool retDelims;
		private bool delimsChanged;
		private int maxDelimCodePoint;
		private bool hasSurrogates = false;

		private int[] delimiterCodePoints;

		private const int MIN_SUPPLEMENTARY_CODE_POINT = 0x010000;
		private const char MIN_HIGH_SURROGATE = '\uD800';
		private const char MAX_HIGH_SURROGATE = '\uDBFF';
		private const char MIN_LOW_SURROGATE  = '\uDC00';
		private const char MAX_LOW_SURROGATE  = '\uDFFF';
		private const char MIN_SURROGATE = MIN_HIGH_SURROGATE;
		private const char MAX_SURROGATE = MAX_LOW_SURROGATE;

		private static int charCount(int codePoint) {
			return codePoint >= MIN_SUPPLEMENTARY_CODE_POINT? 2 : 1;
		}

		private static bool isHighSurrogate(char ch) {
			return ch >= MIN_HIGH_SURROGATE && ch <= MAX_HIGH_SURROGATE;
		}

		private static bool isLowSurrogate(char ch) {
			return ch >= MIN_LOW_SURROGATE && ch <= MAX_LOW_SURROGATE;
		}

		private static int toCodePoint(char high, char low) {
			return ((high - MIN_HIGH_SURROGATE) << 10)
				+ (low - MIN_LOW_SURROGATE) + MIN_SUPPLEMENTARY_CODE_POINT;
		}

		private static int codePointAt(String a, int index, int limit) {
			char c1 = a[index++];
			if (isHighSurrogate(c1)) {
				if (index < limit) {
					char c2 = a[index];
					if (isLowSurrogate(c2)) {
						return toCodePoint(c1, c2);
					}
				}
			}
			return c1;
		}

		private void setMaxDelimCodePoint() 
		{
			if (delimiters == null) {
				maxDelimCodePoint = 0;
				return;
			}

			int m = 0;
			int c;
			int count = 0;
			for (int i = 0; i < delimiters.Length; i += charCount(c)) {
				c = delimiters[i];
				if (c >= MIN_HIGH_SURROGATE && c <= MAX_LOW_SURROGATE) {
					c = codePointAt(delimiters, i, delimiters.Length);
					hasSurrogates = true;
				}
				if (m < c)
					m = c;
				count++;
			}
			maxDelimCodePoint = m;

			if (hasSurrogates) {
				delimiterCodePoints = new int[count];
				for (int i = 0, j = 0; i < count; i++, j += charCount(c)) {
					c = codePointAt(delimiters, j, delimiters.Length);
					delimiterCodePoints[i] = c;
				}
			}
		}

		public JavaTokenizer(String str, String delim, bool returnDelims) {
			init(str, delim, returnDelims);
		}

		private void init(String str, String delim, bool returnDelims) {
			currentPosition = 0;
			newPosition = -1;
			delimsChanged = false;
			this.str = str;
			maxPosition = str.Length;
			delimiters = delim;
			retDelims = returnDelims;
			setMaxDelimCodePoint();
		}

		public JavaTokenizer(String str, String delim) {
			init(str, delim, false);
		}

		public JavaTokenizer(String str) {
			init(str, " \t\n\r\f", false);
		}

		private int skipDelimiters(int startPos) {
			if (delimiters == null)
				throw new NullReferenceException();

			int position = startPos;
			while (!retDelims && position < maxPosition) {
				if (!hasSurrogates) {
					char c = str[position];
					if (c > maxDelimCodePoint || delimiters.IndexOf(c) < 0)
						break;
					position++;
				} 
				else {
					int c = codePointAt(str, position, str.Length);
					if (c > maxDelimCodePoint || !isDelimiter(c)) 
						break;
					position += charCount(c);
				}
			}
			return position;
		}

		private int scanToken(int startPos) {
			int position = startPos;
			while (position < maxPosition) {
				if (!hasSurrogates) {
					char c = str[position];
					if (c <= maxDelimCodePoint && delimiters.IndexOf(c) >= 0)
						break;
					position++;
				} 
				else {
					int c = codePointAt(str, position, str.Length);
					if (c <= maxDelimCodePoint && isDelimiter(c))
						break;
					position += charCount(c);
				}
			}
			if (retDelims && startPos == position) {
				if (!hasSurrogates) {
					char c = str[position];
					if (c <= maxDelimCodePoint && delimiters.IndexOf(c) >= 0)
						position++;
				} 
				else {
					int c = codePointAt(str, position, str.Length);
					if (c <= maxDelimCodePoint && isDelimiter(c))
						position += charCount(c);
				}
			}
			return position;
		}

		private bool isDelimiter(int codePoint) {
			for (int i = 0; i < delimiterCodePoints.Length; i++) {
				if (delimiterCodePoints[i] == codePoint) {
					return true;
				}
			}
			return false;
		}

		public bool hasMoreTokens() {
			newPosition = skipDelimiters(currentPosition);
			return (newPosition < maxPosition);
		}

		public String nextToken() {
			currentPosition = (newPosition >= 0 && !delimsChanged) ?  
				newPosition : skipDelimiters(currentPosition);

			/* Reset these anyway */
			delimsChanged = false;
			newPosition = -1;

			if (currentPosition >= maxPosition)
				throw new ArgumentOutOfRangeException();
			int start = currentPosition;
			currentPosition = scanToken(currentPosition);
			return str.Substring(start, currentPosition - start);
		}

		public String nextToken(String delim) {
			delimiters = delim;
			delimsChanged = true;
			setMaxDelimCodePoint();
			return nextToken();
		}

		public bool hasMoreElements() {
			return hasMoreTokens();
		}

		public Object nextElement() {
			return nextToken();
		}

		public int countTokens() {
			int count = 0;
			int currpos = currentPosition;
			while (currpos < maxPosition) {
				currpos = skipDelimiters(currpos);
				if (currpos >= maxPosition)
					break;
				currpos = scanToken(currpos);
				count++;
			}
			return count;
		}
	}
}
